#
#     Function file: 4.3.functions.s
#
#     Example 4.3: Serially correlated errors
#
#     =======================================================
#
#     Compute B-spline base matrix
#
bspline <- function(x, xl, xr, ndx, bdeg){
dx <- (xr-xl)/ndx
knots <- seq(xl-bdeg*dx, xr+bdeg*dx, by=dx)
B <- spline.des(knots,x,bdeg+1,0*x)$design
B
}
#
#     Compute matrices for Pspline
#
Pspline.setup <- function(x, xl, xr, ndx, bdeg, pord)
{
B. <- bspline(x, xl, xr, ndx, bdeg)
D. <- diag(ncol(B.))
for(k in 1:pord) D. <- diff(D.)
Z. <- t(D.) %*% solve(D. %*% t(D.))
X. <- NULL
for(k in 1:pord) X. <- cbind(X., (1:ncol(D.))^(k - 1))
tX <- B. %*% X.
return(B., Z., tX, D.)
}
#
#     Determinant function
#
det <- function(M) prod(svd(M, nu=0, nv=0)$d)
#
#    QR function to compute R00, R11.1 and cm1 in 
#    Pinheiro and Bates p 76, equation (2.23) 
#
QR <- function(X., Z., Y., Lambda.)
{
  p <- ncol(X.)
  q <- ncol(Z.)
  ytil <- c(Y., rep(0, q))
  Xtil <- rbind(X., matrix(0, nrow = q, ncol = ncol(X.)))
  Ztil <- rbind(Z., sqrt(Lambda.) * diag(q))
  Ztil.qr <- qr(Ztil)
  R11.1 <- qr.R(Ztil.qr)[1:q, 1:q]
  QtXy <- qr.qty(Ztil.qr, cbind(Xtil, ytil))
  R00.1 <- QtXy[(q + 1):nrow(QtXy), 1:(ncol(QtXy) - 1)]
  R10 <- QtXy[1:q, 1:p]
  c0.1 <- QtXy[(q + 1):nrow(QtXy), ncol(QtXy)]
  c1 <- QtXy[1:q, ncol(QtXy)]
  Rc.qr <- qr(cbind(R00.1, c0.1))
  R <- qr.R(Rc.qr)
  k <- ncol(R) - 1
  R00 <- R[1:k, 1:k]
  c0 <- R[1:k, k + 1]
  cm1 <- R[(k + 1):nrow(R), (k + 1)]
  return(R11.1, R00, cm1, c0, c1, R10)
}
 #
 #     Evaluate the profile REML via QR
 #     Pinheiro and Bates, p76
 #
 P.Reml.QR.cor <- function(X., Z., Y., Lambda., rho1, rho2, order){
   n <- length(Y.)
   rho1 <- min(rho1, 0.9999999)
   Lambda. <- abs(Lambda.)
   if(order == 2) {
     W <- ar2.matrix(n, rho1, rho2)
   }
   else if(order == 1) {
     W <- ar1.matrix(n, rho1)
   }
   W <- t(chol(W))
   W.inv <- solve(W)
   Z. <- W.inv %*% Z.
   X. <- W.inv %*% X.
   Y. <- W.inv %*% Y.
   qr <- QR(X., Z., Y., Lambda.)
   p <- ncol(X.)
   l1 <- -0.5*(n-p)*log(sum(qr$cm1^2))
   l2 <- -log(abs(det(qr$R00)))
   l3 <- log( det(sqrt(Lambda.)*diag(ncol(Z.)))/det(abs(qr$R11.1)) )
   l4 <- sum(log(diag(W.inv)))
   return(l1 + l2 + l3+l4)
 }
 # 
 #  Matrix for AR1
 #
 ar1.matrix<-function(n, rho1){
   X <- diag(n)
   for(i in 2:n) X[i, ] <- X[i, ] + rho1 * X[i - 1, ]
   X <- X + t(X) - diag(diag(X))
   X
}
#
#  Matrix for AR2
#
ar2.matrix <- function(n, rho1, rho2){
  X <- diag(n)
  X <- rho1/(rho1 * (1 - rho2)) * X
  X[1, 1] <- 1
  X[2,  ] <- X[2,  ] + rho1/(1 - rho2) * X[1,  ]
  for(i in 3:n) {
    X[i, ] <- rho1 * X[i - 1, ] + rho2 * X[i - 2, ] + X[i, ]
    X[i - 1, i - 1] <- 1
  }
  X[n, n] <- 1
  X <- X + t(X) - diag(diag(X))
  X
}
#
#    Compute fitted values and sigma^2
#
Fit.Val.cor <- function(X., Z., Y., Lambda., rho1, rho2, order){
   n <- nrow(X.)
   p <- ncol(X.)
   if(order == 2) {
     W <- ar2.matrix(n, rho1, rho2)
   }
   else if(order == 1) {
     W <- ar1.matrix(n, rho1)
   }
   W <- t(chol(W))
   W.inv <- solve(W)
   Z. <- W.inv %*% Z.
   X. <- W.inv %*% X.
   Y. <- W.inv %*% Y.
   qr <- QR(X., Z., Y., Lambda.)
   beta <- solve(qr$R00) %*% qr$c0
   b <- solve(qr$R11.1) %*% (qr$c1 - qr$R10 %*% beta)
   fit <- W %*% (X. %*% beta + Z. %*% b)
   sigma.sq <- sum(qr$cm1^2)/(n-p)
   return(fit, sigma.sq)
 }
#
#    Compute confidence intervals
#
ci.cor <- function(B., y, lambda, rho1, rho2, D., sigma, order){
  n <- nrow(X.)
  if(order == 2) {
     W <- ar2.matrix(n, rho1, rho2)
  }
  else if(order == 1) {
     W <- ar1.matrix(n, rho1)
  }
  V.inv <- solve(W)
  Hat.H <- B. %*% solve(t(B.) %*% V.inv %*% B. + 
		lambda * t(D.) %*% D.) %*% t(B.) %*% V.inv
  ci <- 1.96 * sqrt(sigma) * sqrt(diag(Hat.H %*% W %*% t(Hat.H)))
  return(ci)
}
